/*----------------------- BEGIN small_rates.cu BEGIN ----------------------*/
/* @file small_rates.cu                                                    */
/* @author jlinford                                                        */
/* @date 2012-06-22 15:44:03.961781                                        */
/* @brief Reaction rate calculation and utility functions                  */
/*                                                                         */
/* Reaction rate calculation and utility functions                         */
/*                                                                         */
/* This file was generated by Kppa: http://www.paratools.com/Kppa          */
/*-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include "small_cu_parameters.h"
#include "small_rates.h"


/*---------------------------------- ARR ----------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
__device__
float ARR(float const a0, float const b0, float const c0, float const temp)
{
    return __fmul_rn(__fmul_rn(a0, __powf(__fdiv_rn(temp,(float)300.0), c0)),
                  __expf(__fdiv_rn(-b0,temp)));
} /* END ARR */


/*---------------------------------- ARR2 ---------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
__device__
float ARR2(float const a0, float const b0, float const temp)
{
    return __fmul_rn(a0, __expf(__fdiv_rn(b0,temp)));
} /* END ARR2 */


/*---------------------------------- EP2 ----------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     a2   None                                                */
/* @param[in]     c2   None                                                */
/* @param[in]     a3   None                                                */
/* @param[in]     c3   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
__device__
float EP2(float const a0, float const c0, float const a2, float const c2,
        float const a3, float const c3, float const temp)
{
    float k0 = __fmul_rn(a0, __expf(__fdiv_rn(-c0,temp)));
    float k2 = __fmul_rn(a2, __expf(__fdiv_rn(-c2,temp)));
    float k3 = __fmul_rn(a3, __expf(__fdiv_rn(-c3,temp))) * (float)1.0e6*CFACTOR;
    return __fadd_rn(k0, __fdiv_rn(k3,
                __fadd_rn((float)1.0, __fdiv_rn(k3,k2))));
} /* END EP2 */


/*---------------------------------- EP3 ----------------------------------*/
/* @param[in]     a1   None                                                */
/* @param[in]     c1   None                                                */
/* @param[in]     a2   None                                                */
/* @param[in]     c2   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
__device__
float EP3(float const a1, float const c1, float const a2, float const c2,
        float const temp)
{
    float k1 = __fmul_rn(a1, __expf(__fdiv_rn(-c1,temp)));
    float k2 = __fmul_rn(a2, __expf(__fdiv_rn(-c2,temp)));
    return __fadd_rn(k1, __fmul_rn(k2, (float)1.0e6*CFACTOR));
} /* END EP3 */


/*---------------------------------- FALL ---------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     a1   None                                                */
/* @param[in]     b1   None                                                */
/* @param[in]     c1   None                                                */
/* @param[in]     cf   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
__device__
float FALL(float const a0, float const b0, float const c0, float const a1,
        float const b1, float const c1, float const cf, float const temp)
{
    /* Accuracy trumps precision in these calculations */
    float k0 = a0 * pow(temp/(float)300.0, c0) * exp(-b0/temp) * (float)1.0e6*CFACTOR; 
    float k1 = k0 / (a1 * pow(temp/(float)300.0, c1) * exp(-b1/temp));
    return (k0/((float)1.0+k1)) * pow(cf, (float)1.0/((float)1.0+pow(log10(k1),(float)2.0)));
} /* END FALL */


/*-------------------------------- Sunlight -------------------------------*/
/* Calculates sunlight intensity in the range [0,1] as a function of time. */
/* Modify this routine to get the correct sunlight values for your model.  */
/*                                                                         */
/* @param[in]     time Integration time                                    */
/* @param[in]     idx  Current grid cell index                             */
/*-------------------------------------------------------------------------*/
__device__
float Sunlight(float const time, unsigned long const idx)
{
    int daysec = 24 * 3600;     /* Seconds per day */
    float sunrise = 5.5*3600;   /* 5:30 local time */
    float sunset = 19.5*3600;   /* 7:30 local time */

    float daily = time - ((int)time / daysec) * daysec;
    float tmp;

    /* Estimate sunlight intensity in the range [0,1] */
    if ((daily >= sunrise) && (daily <= sunset)) {
        tmp = __fdiv_rn(2.0*daily - sunrise-sunset, sunset-sunrise);
        tmp = (tmp > 0) ? tmp * tmp : -tmp * tmp;
        tmp = 0.5 * (1.0 + cospi(tmp));
    } else {
        tmp = 0.0;
    }
    return tmp;
} /* END Sunlight */



/*------------------------------ Temperature ------------------------------*/
/* Calculates temperature (kelvin) as a function of time.                  */
/* Modify this routine to get the correct temperature values for your model. */
/*                                                                         */
/* @param[in]     time Integration time                                    */
/* @param[in]     idx  Current grid cell index                             */
/*-------------------------------------------------------------------------*/
__device__
float Temperature(float const time, unsigned long const idx)
{
    float mintemp = 280;    /* 280 Kelvin ~= 44 Fahrenheit */
    float maxtemp = 300;    /* 300 Kelvin ~= 80 Fahrenheit */
    float tmp;

    /* Estimate temperature cycling from mintemp to maxtemp */
    tmp = sinpi(__fdiv_rn(time,24*3600));
    if (tmp < 0) {
        tmp = mintemp - tmp * (maxtemp-mintemp);
    } else {
        tmp = mintemp + tmp * (maxtemp-mintemp);
    }
    return tmp;
} /* END Temperature */



/*-------------------------------- d_Rates --------------------------------*/
/* CUDA kernel for Rates                                                   */
/*                                                                         */
/* @param[in]     ncells32 A multiple of 32 grid cells                     */
/* @param[in]     time     Integration time                                */
/* @param[in]     idx      Current grid cell index                         */
/* @param[out]    rct      Reaction rates                                  */
/*-------------------------------------------------------------------------*/
__global__
void d_Rates(unsigned long const ncells32, float const time, unsigned long
        const idx, float rct[])
{
    /* Sunlight intensity: 0 to 1 inclusive (uppercase for KPP compatibility) */
    float SUN;
    /* Temperature in kelvin  (uppercase for KPP compatibility) */
    float TEMP;

    SUN = Sunlight(time, idx);
    TEMP = Temperature(time, idx);

    size_t tidx = blockDim.x*blockIdx.x + threadIdx.x;
    if(tidx < ncells32) {
        rct += tidx;
        rct[0] = (2.643E-10) * SUN*SUN*SUN;
        rct[ncells32] = 8.018e-17;
        rct[2*ncells32] = (6.120E-04) * SUN;
        rct[3*ncells32] = 1.576e-15;
        rct[4*ncells32] = (1.070E-03) * SUN*SUN;
        rct[5*ncells32] = 7.11e-11;
        rct[6*ncells32] = 1.2e-10;
        rct[7*ncells32] = 6.062e-15;
        rct[8*ncells32] = 1.069e-11;
        rct[9*ncells32] = (1.289E-02) * SUN;
    }
} /* END d_Rates */


/*--------------------------------- Rates ---------------------------------*/
/* Calculates reaction rate coefficients                                   */
/*                                                                         */
/* @param[in]     ncells32 A multiple of 32 grid cells                     */
/* @param[in]     time     Integration time                                */
/* @param[in]     idx      Current grid cell index                         */
/* @param[out]    d_rct    Reaction rates in device memory                 */
/*-------------------------------------------------------------------------*/
void Rates(unsigned long const ncells32, float const time, unsigned long
        const idx, float d_rct[])
{
    size_t nBlocks = ((ncells32 + 127) & ~127) >> 7;
    size_t nThreads = 128;
    d_Rates<<<nBlocks, nThreads>>>(ncells32, time, idx, d_rct);
} /* END Rates */


/*------------------------- END small_rates.h END -------------------------*/
